package confext

import (
	"errors"
	"fmt"
	"gogs.mikescher.com/BlackForestBytes/goext/timeext"
	"math/bits"
	"os"
	"reflect"
	"strconv"
	"time"
)

// ApplyEnvOverrides overrides field values from environment variables
//
// fields must be tagged with `env:"env_key"`
func ApplyEnvOverrides[T any](c *T) error {

	rval := reflect.ValueOf(c).Elem()
	rtyp := rval.Type()

	for i := 0; i < rtyp.NumField(); i++ {

		rsfield := rtyp.Field(i)
		rvfield := rval.Field(i)

		envkey := rsfield.Tag.Get("env")
		if envkey == "" {
			continue
		}

		envval, efound := os.LookupEnv(envkey)
		if !efound {
			continue
		}

		if rvfield.Type() == reflect.TypeOf("") {

			rvfield.Set(reflect.ValueOf(envval))

			fmt.Printf("[CONF] Overwrite config '%s' with '%s'\n", envkey, envval)

		} else if rvfield.Type() == reflect.TypeOf(int(0)) {

			envint, err := strconv.ParseInt(envval, 10, bits.UintSize)
			if err != nil {
				return errors.New(fmt.Sprintf("Failed to parse env-config variable '%s' to int (value := '%s')", envkey, envval))
			}

			rvfield.Set(reflect.ValueOf(int(envint)))

			fmt.Printf("[CONF] Overwrite config '%s' with '%s'\n", envkey, envval)

		} else if rvfield.Type() == reflect.TypeOf(int64(0)) {

			envint, err := strconv.ParseInt(envval, 10, 64)
			if err != nil {
				return errors.New(fmt.Sprintf("Failed to parse env-config variable '%s' to int64 (value := '%s')", envkey, envval))
			}

			rvfield.Set(reflect.ValueOf(int64(envint)))

			fmt.Printf("[CONF] Overwrite config '%s' with '%s'\n", envkey, envval)

		} else if rvfield.Type() == reflect.TypeOf(int32(0)) {

			envint, err := strconv.ParseInt(envval, 10, 32)
			if err != nil {
				return errors.New(fmt.Sprintf("Failed to parse env-config variable '%s' to int32 (value := '%s')", envkey, envval))
			}

			rvfield.Set(reflect.ValueOf(int32(envint)))

			fmt.Printf("[CONF] Overwrite config '%s' with '%s'\n", envkey, envval)

		} else if rvfield.Type() == reflect.TypeOf(int8(0)) {

			envint, err := strconv.ParseInt(envval, 10, 8)
			if err != nil {
				return errors.New(fmt.Sprintf("Failed to parse env-config variable '%s' to int32 (value := '%s')", envkey, envval))
			}

			rvfield.Set(reflect.ValueOf(int8(envint)))

			fmt.Printf("[CONF] Overwrite config '%s' with '%s'\n", envkey, envval)

		} else if rvfield.Type() == reflect.TypeOf(time.Duration(0)) {

			dur, err := timeext.ParseDurationShortString(envval)
			if err != nil {
				return errors.New(fmt.Sprintf("Failed to parse env-config variable '%s' to duration (value := '%s')", envkey, envval))
			}

			rvfield.Set(reflect.ValueOf(dur))

			fmt.Printf("[CONF] Overwrite config '%s' with '%s'\n", envkey, dur.String())

		} else if rvfield.Type() == reflect.TypeOf(time.UnixMilli(0)) {

			tim, err := time.Parse(time.RFC3339Nano, envval)
			if err != nil {
				return errors.New(fmt.Sprintf("Failed to parse env-config variable '%s' to time.time (value := '%s')", envkey, envval))
			}

			rvfield.Set(reflect.ValueOf(tim))

			fmt.Printf("[CONF] Overwrite config '%s' with '%s'\n", envkey, tim.String())

		} else if rvfield.Type().ConvertibleTo(reflect.TypeOf(int(0))) {

			envint, err := strconv.ParseInt(envval, 10, 8)
			if err != nil {
				return errors.New(fmt.Sprintf("Failed to parse env-config variable '%s' to <%s, ,int> (value := '%s')", rvfield.Type().Name(), envkey, envval))
			}

			envcvl := reflect.ValueOf(envint).Convert(rvfield.Type())

			rvfield.Set(envcvl)

			fmt.Printf("[CONF] Overwrite config '%s' with '%v'\n", envkey, envcvl.Interface())

		} else if rvfield.Type().ConvertibleTo(reflect.TypeOf("")) {

			envcvl := reflect.ValueOf(envval).Convert(rvfield.Type())

			rvfield.Set(envcvl)

			fmt.Printf("[CONF] Overwrite config '%s' with '%v'\n", envkey, envcvl.Interface())

		} else {
			return errors.New(fmt.Sprintf("Unknown kind/type in config: [ %s | %s ]", rvfield.Kind().String(), rvfield.Type().String()))
		}
	}

	return nil
}
